////////////////////////////////////////////////////////////////////// 
// KinshipEmp.cpp 
// (c) 2012-2013 Shuang Feng, Dajiang Liu, Goncalo Abecasis
// 
// This file is distributed as part of the RareMetalWorker source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile RareMetalWorker.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Wednesday November 28, 2012
// 

#include "KinshipEmp.h"
#include "MathMatrix.h"
#include "PedigreeAlleleFreq.h"
#include "OutputKin.h"
#include "PreMeta.h"
#include <string>
#include <cstring>
#include <iostream>
#include <stdlib.h>
#include <vector>
#include "math.h"
#include <algorithm>

double KinshipEmp::q = 0.05;
double KinshipEmp::miss = 0.05;

void KinshipEmp::Setup(Pedigree & ped,IntArray * pPheno, IntArray & genotypedSamplePED, FILE * log)
{
    printf("Calculating empirical kinship matrix...\n");
    fprintf(log,"Calculating empirical kinship matrix...\n");

    int N=0;
    int total_n = genotypedSamplePED.Length(); 

    allPairs.Dimension(total_n, total_n);
    allPairs.Zero();
    Vector X;
    X.Dimension(total_n);

    for(int i=0;i<ped.markerCount;i++) 
    {
       int alleleCount = CountAlleles(ped, i);
       //if no one is genotyped, mornomorphic, or marker is not diallelic
       if(alleleCount<=1 || alleleCount>2)
	  continue;

       //calculate allele frequencies
       bool tmp = EstimateFrequencies(ped,i,FREQ_ALL); 
       Vector freq = ped.GetMarkerInfo(i)->freq;
       double maf = 0.0;
       int minor;
       if(freq[1]<freq[2]) 
       {
	  minor = 1;
	  maf = freq[1];
       }
       else
       {
	  minor=2;
	  maf = freq[2];
       }
       if(maf<q)
	  continue;

       //calculate the missing rate
       int nmiss=0;
       for(int idv=0;idv<total_n;idv++)
       {
	  int p = genotypedSamplePED[idv];
	  if(!ped[p].isGenotyped(i))
	     nmiss++;
       }
       if(nmiss > total_n*miss)
	  continue;

       N++; //useful marker count ++

       double mean=0.0;
       double variance = 0.0;
       double mean_sqr = 0.0;
       //now fill in the genotypes
       int n_=0;
       //only generate matrix for genotyped individuals
       for(int p =0;p<total_n;p++)
       {
	  if(!ped[genotypedSamplePED[p]].isGenotyped(i))
	  {
	     X[p] = -1.0;
	  }
	  else if(ped[genotypedSamplePED[p]].markers[i].isHeterozygous())
	  {
	     X[p]=1.0;
	     mean += X[p];
	     mean_sqr += X[p]*X[p];
	     n_++;
	  }
	  else if(ped[genotypedSamplePED[p]].markers[i].isHomozygous() && ped[genotypedSamplePED[p]].markers[i].hasAllele(minor))
	  {
	     X[p]=2.0;
	     mean += X[p];
	     mean_sqr += X[p]*X[p];
	     n_++;
	  }
	  else 
	  {
	     X[p]=0.0;
	     mean += X[p];
	     mean_sqr += X[p]*X[p];
	     n_++;
	  }
       }

       mean /= n_;
       variance = mean_sqr/n_ - mean*mean;

       for(int p=0;p<total_n;p++)
       {
	  if(X[p]==-1.0)
	     X[p] =0.0;
	  else
	     X[p] -= mean;
       }

       for(int r=0;r<total_n;r++)
       {
	  for(int c=r;c<total_n;c++)
	  {
	     allPairs[r][c] += X[r]*X[c]/variance;
	  }
       }
    }

    for(int i=0;i<total_n;i++)
       for(int j=i+1;j<total_n;j++)
	  allPairs[j][i] = allPairs[i][j];

    allPairs.Multiply(1.0/N);

    printf("Finished calculating empirical kinship matrix. \n ");
    fprintf(log,"Finished calculating empirical kinship matrix. \n ");
    if(OutputKin::outputKin) 
    {
       printf("Saving empirical kinship matrix...\n");
       fprintf(log,"Saving empirical kinship matrix...\n");
       FILE * file = NULL;
       String filename;
       if(PreMeta::outputFile != "")
       {
	  filename.printf("%s.Empirical.Kinship",(const char *) PreMeta::outputFile);
       }
       else
       {
	  filename.printf("Empirical.Kinship");
       }
       file = fopen(filename, "wt");
       for(int p=0;p<total_n;p++)
       {  
	  fprintf(file,"%s ",ped[genotypedSamplePED[p]].pid.c_str());
       }
       fprintf(file,"\n");

       for(int r=0;r<allPairs.rows;r++)
       {
	  for(int c=0;c<allPairs.cols;c++)
	  {
	     fprintf(file,"%g ",allPairs[r][c]);
	  }
	  fprintf(file,"\n");
       }
       fclose(file);
    }
}

void KinshipEmp::SetupVCF(Pedigree & ped, IntArray * pPheno,IntArray & genotypedSampleVCF,FILE * log)
{
   printf("Calculating empirical kinship matrix...\n");
   fprintf(log,"Calculating empirical kinship matrix...\n");
   VcfFileReader reader;
   VcfHeader header;
   VcfRecord record;
   reader.open(PreMeta::vcfInput,header);

   int N=0;
   int total_n = genotypedSampleVCF.Length();
   allPairs.Dimension(total_n,total_n);
   allPairs.Zero();

   Vector genotype;
   genotype.Dimension(total_n);

   if(PreMeta::dosage)
   {
      while(reader.readRecord(record))
      {
	 //step1: decide if this marker should be included
	 int n=0,nmiss=0;
	 double maf=0.0;
	 VcfRecordGenotype & genoInfo = record.getGenotypeInfo();
	 for(int i=0;i<total_n;i++)
	 {
	    int s = genotypedSampleVCF[i];
	    const std::string * geno = genoInfo.getString("DS",s);

	    if(*geno == ".")
	    {
	       nmiss++;
	       continue;
	    }
	    n++;
	    float dose = atof((*geno).c_str());

	    //Calculate overall maf
	    maf += dose;
	 }
	 //calculate final maf
	 maf /= 2.0*n;
	 if(maf>0.5) 
	    maf = 1.0-maf;

	 if(nmiss > total_n*miss || maf<q)
	    continue;
	 N++;

	 //step2: if yes, then loop through all samples to get genotype. 
	 genotype.Zero();
	 double mean = 0.0, mean_sqr= 0.0;
	 int n_ = 0;
	 for (int p = 0; p < total_n; p++)
	 {
	    int s = genotypedSampleVCF[p];
	    //fill in genotype vector for this sample
	    const std::string * geno = genoInfo.getString("DS",s);
	    if(*geno == ".")
	       genotype[p] = _NAN_;
	    else
	    {
	       genotype[p] = atof((*geno).c_str());
	       mean += genotype[p];
	       n_++;
	       mean_sqr += genotype[p]*genotype[p];
	    }
	 }

	 //step 3: standardize genotype vector
	 mean /= n_;
	 double variance = mean_sqr/n_ - mean*mean;
	 for(int i=0;i<genotype.Length();i++)
	 {
	    if(genotype[i]==_NAN_)
	       //make sure if genotype is missing, then it is not contributing
	       genotype[i]=0.0;
	    else
	       genotype[i] -= mean;
	 }
	 //add up to kinship
	 for(int r=0;r<total_n;r++)
	 {
	    for(int c=r;c<total_n;c++)
	    {
	       allPairs[r][c] += genotype[r]*genotype[c]/variance;
	    }
	 }
      }

      for(int i=0;i<total_n;i++)
	 for(int j=i+1;j<total_n;j++)
	    allPairs[j][i] = allPairs[i][j];

      allPairs.Multiply(1.0/N);
   }
   else
   {
      IntArray counts;
      counts.Dimension(2);
      while(reader.readRecord(record))
      {
	 //step1: decide if this marker should be included
	 double maf=0.0;
	 int nmiss =0;
	 counts.Zero();
	 //	 fcounts.Zero();
	 bool skip = false;
	 for(int s=0;s<total_n;s++)
	 {
	    int i= genotypedSampleVCF[s];
	    bool miss_stat = false;
	    int numGTs = record.getNumGTs(i);
	    for(int j = 0; j < numGTs; j++)
	    {
	       int a = record.getGT(i,j);
	       if(a==VcfGenotypeSample::MISSING_GT)
	       {
		  miss_stat=true;
		  nmiss++;
		  break;
	       }
	       if(a>1)
	       {
		  skip = true;
		  //if marker is not biallelic, skip this marker
		  goto skip_to_point;
	       }
	       counts[a]++;
	    }
	    if(miss_stat)
	    {
	       //if genotype is missing, skip this sample.
	       continue;
	    }
	 }

skip_to_point: 
	 if(skip)
	    continue;

	 if(counts[0]==0 || counts[1]==0)
	    continue;

	 //Calculate maf
	 maf = (double)counts[1]/(counts[0]+counts[1]);
	 if(maf>0.5) 
	    maf = 1.0-maf;

	 if(nmiss > total_n*miss || maf<q)
	 {
	    continue;
	 }

	 N++;
	 //printf("%d\n",N);

	 //step2: if yes, then loop through all samples to get genotype. 

	 int n_=0;
	 double mean = 0.0,mean_sqr=0.0;
	 genotype.Zero();
	 for (int i = 0;i < total_n; i++)
	 {
	    int s = genotypedSampleVCF[i];
	    int numGTs = record.getNumGTs(s);
	    bool skip=false;
	    for(int j = 0; j < numGTs; j++)
	    {
	       int a = record.getGT(s,j);
	       if(a==VcfGenotypeSample::MISSING_GT)
	       {
		  genotype[i] = _NAN_;
		  skip=true;
		  break;
	       }
	       genotype[i] += a;
	    }
	    if(skip)
	       continue;

	    mean += genotype[i];
	    n_++;
	    mean_sqr += genotype[i]*genotype[i];
	 }

	 //step 3: standardize genotype vector
	 mean /= n_;
	 double variance = mean_sqr/n_ - mean*mean;
	 for(int i=0;i<genotype.Length();i++)
	 {
	    if(genotype[i]==_NAN_)
	    {
	       //make sure if genotype is missing, then it is not contributing
	       genotype[i]=0.0;
	    }
	    else
	    {
	       genotype[i] -= mean;
	    }
	 }
	 //add up to kinship
	 //	 printf("%s %s variance =%g\n",record.getChromStr(),record.get1BasedPosition(),variance);
	 for(int r=0;r<total_n;r++)
	 {
	    for(int c=r;c<total_n;c++)
	    {
	       allPairs[r][c] += genotype[r]*genotype[c]/variance;
	    }
	 }
      }

      for(int i=0;i<total_n;i++)
	 for(int j=i+1;j<total_n;j++)
	    allPairs[j][i] = allPairs[i][j];

      if(N==0)
	 error("ERROR! No qualified marker to use for kinship calculation.\n");

      allPairs.Multiply(1.0/N);
   }

   if(OutputKin::outputKin)
   {
      printf("Writing kinship matrix ...\n");
      fprintf(log,"Writing kinship matrix ...\n");
      FILE * file = NULL;
      String filename;
      if(PreMeta::outputFile != "")
	 filename.printf("%s.Empirical.Kinship",(const char *) PreMeta::outputFile);
      else
	 filename.printf("Empirical.Kinship");
      file = fopen(filename, "wt");
      for (int i = 0; i < total_n; i++)
      {
	 const char * sample = header.getSampleName(genotypedSampleVCF[i]);
	 fprintf(file,"%s ",sample);
      }
      fprintf(file,"\n");
      for(int r=0;r<allPairs.rows;r++)
      {
	 for(int c=0;c<allPairs.cols;c++)
	 {
	    fprintf(file,"%g ",allPairs[r][c]);
	 }
	 fprintf(file,"\n");
      }
      fclose(file);
   }
}


